from tuya_iot import TuyaOpenAPI, AuthType, TuyaDeviceManager, TuyaOpenMQ

from .device import GosundDevice
from .exceptions import assert_response_success, GosundException
from .utils import cache_response

class Gosund(object):

    def __init__(self, username, password, access_id, access_key, country=1,
            endpoint='https://openapi.tuyaus.com'):
        self.api = TuyaOpenAPI(endpoint, access_id, access_key,
                auth_type=AuthType.CUSTOM)
        resp = self.api.connect(username, password, country)
        assert_response_success('connect to api', resp)
        self.manager = TuyaDeviceManager(self.api, TuyaOpenMQ(self.api))
        self._known_devices = {}

    def get_device(self, device_id):
        resp = self.manager.get_device_functions(device_id)
        assert_response_success('get device', resp)
        self._add_known_device(device_id)
        return GosundDevice.from_response(resp, device_id, self)

    def get_device_status(self, device_id):
        status = self.get_device_statuses().get(device_id)
        if status is None:
            raise GosundException(
                f'unable to find status for device with id "{device_id}"')
        return status

    @cache_response(seconds=60)
    def get_device_statuses(self):
        # limit 20 device_ids per api call
        resp = self.manager.get_device_list_status(self._known_devices)
        assert_response_success('get device statuses', resp)
        return {device['id']: device['status'] for device in resp.get('result', [])}

    def _add_known_device(self, device_id):
        self._known_devices[device_id] = True
        self._clear_statuses_cache()

    def _remove_known_device(self, device_id):
        del self._known_devices[device_id]
        self._clear_statuses_cache()

    def _clear_statuses_cache(self):
        self.get_device_statuses.clear_cache()

    def send_commands(self, device_id, commands):
        resp = self.manager.send_commands(device_id, commands)
        codes = [f'{cmd["code"]}:{cmd["value"]}' for cmd in commands]
        assert_response_success(f'send {codes} command to device', resp)
        self._clear_statuses_cache()
        return resp



import functools
import time

class _timed_cache(object):

    def __init__(self):
        self.reset()

    def set(self, last_call, value):
        self.last_call = last_call
        self.value = value

    def reset(self):
        self.last_call = 0
        self.value = None

def cache_response(hours=0, minutes=0, seconds=0):
    seconds += 60 * 60 * hours + 60 * minutes

    class _cache_response(object):

        def __init__(self, fn):
            # wraps class instance methods only
            self.fn = fn
            self.cache_key = f'_{fn.__name__}_cache'

        def call(self, instance, *args, **kwargs):
            now = time.time()
            cache = getattr(instance, self.cache_key)
            if now - cache.last_call < seconds:
                return cache.value
            cache.set(now, self.fn(instance, *args, **kwargs))
            return cache.value

        def __get__(self, instance, owner):
            @functools.wraps(self.fn)
            def _call(*args, **kwargs):
                return self.call(instance, *args, **kwargs)
            if not hasattr(instance, self.cache_key):
                setattr(instance, self.cache_key, _timed_cache())
            _call.clear_cache = getattr(instance, self.cache_key).reset
            return _call

    return _cache_response



import logging

from .exceptions import GosundException

logger = logging.getLogger('gosundpy')

class GosundDevice(object):

    @staticmethod
    def from_response(resp, device_id, gosund):
        category = resp['result']['category']
        if category == 'cz':
            cls = GosundSwitchDevice
        elif category == 'dj':
            cls = GosundLightBulbDevice
        elif category == 'wsdcg':
            cls = GosundTempuratureHumiditySensorDevice
        elif category == 'ldcg':
            cls = GosundLightSensorDevice
        elif category == 'pir':
            cls = GosundMotionSensorDevice
        elif category == 'mcs':
            cls = GosundContactSensorDevice
        else:
            cls = GosundDevice
        return cls(device_id, gosund)

    def __init__(self, device_id, gosund):
        self.device_id = device_id
        self.gosund = gosund

    def get_status(self):
        return self.gosund.get_device_status(self.device_id)

    def get_status_value(self, code):
        for status in self.get_status():
            if status.get('code') == code:
                return status.get('value')
        raise GosundException(
                f'code "{code}" not found for device id "{self.device_id}"')

    def send_commands(self, commands):
        return self.gosund.send_commands(self.device_id, commands)

    def __str__(self):
        return f'<{self.__class__.__name__} device_id={self.device_id}>'
    __repr__ = __str__

class DeviceOnOffMixin(object):

    def turn_on(self):
        logger.debug('turning on device %s', self.device_id)
        commands = [{'code': self.on_off_code, 'value': True}]
        return self.send_commands(commands)

    def turn_off(self):
        logger.debug('turning off device %s', self.device_id)
        commands = [{'code': self.on_off_code, 'value': False}]
        return self.send_commands(commands)

    def is_on(self):
        return self.get_status_value(self.on_off_code)

    def switch(self):
        logger.debug('switching device %s', self.device_id)
        value = not self.is_on()
        commands = [{'code': self.on_off_code, 'value': value}]
        return self.send_commands(commands)

class GosundSwitchDevice(GosundDevice, DeviceOnOffMixin):

    on_off_code = 'switch_1'

class GosundLightBulbDevice(GosundDevice, DeviceOnOffMixin):

    on_off_code = 'switch_led'

class GosundTempuratureHumiditySensorDevice(GosundDevice):

    temperature_code = 'va_temperature'
    humidity_code = 'va_humidity'

    def get_temperature(self, unit='F'):
        val = self.get_status_value(self.temperature_code)
        temp = val / 10
        if unit.upper() == 'F':
            temp = temp * 9/5 + 32
        return temp

    def get_humidity(self):
        return self.get_status_value(self.humidity_code)

class GosundLightSensorDevice(GosundDevice):

    lux_code = 'bright_value'

    def get_lux(self):
        return self.get_status_value(self.lux_code)

class GosundMotionSensorDevice(GosundDevice):

    motion_code = 'pir'

    NO_ONE_STATE = 'none'
    HUMAN_STATE = 'pir'

    def motion_sensed(self):
        return self.get_status_value(self.motion_code) == self.HUMAN_STATE

class GosundContactSensorDevice(GosundDevice):

    contact_code = 'doorcontact_state'

    def is_open(self):
        return self.get_status_value(self.contact_code)



import os

from .device import (GosundDevice, GosundSwitchDevice, GosundLightBulbDevice,
        GosundTempuratureHumiditySensorDevice, GosundLightSensorDevice,
        GosundMotionSensorDevice, GosundContactSensorDevice)
from .exceptions import GosundException
from .gosund import Gosund
from .version import version

__all__ = [
        'GosundDevice',
        'GosundSwitchDevice',
        'GosundLightBulbDevice',
        'GosundTempuratureHumiditySensorDevice',
        'GosundLightSensorDevice',
        'GosundMotionSensorDevice',
        'GosundContactSensorDevice',
        'GosundException',
        'Gosund',
]

__version__ = version
